@@ -1,4 +1,4 @@
-/*	$OpenBSD: config.c,v 1.113 2024/12/13 19:21:03 claudio Exp $ */
+/*	$OpenBSD: config.c,v 1.116 2025/03/26 15:28:13 claudio Exp $ */
 
 /*
  * Copyright (c) 2003, 2004, 2005 Henning Brauer <henning@openbsd.org>
@@ -18,13 +18,13 @@
 
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <arpa/inet.h>
 
 #include <errno.h>
 #include <ifaddrs.h>
 #include <netdb.h>
 #include <stddef.h>
 #include <stdlib.h>
-#include <stdio.h>
 #include <string.h>
 #include <unistd.h>
 
@@ -407,9 +407,7 @@ merge_config(struct bgpd_config *xconf, struct bgpd_config *conf)
 			if (ola->flags & DEFAULT_LISTENER)
 				ola->reconf = RECONF_KEEP;
 	/* else loop over listeners and merge configs */
-	for (nla = TAILQ_FIRST(conf->listen_addrs); nla != NULL; nla = next) {
-		next = TAILQ_NEXT(nla, entry);
-
+	TAILQ_FOREACH_SAFE(nla, conf->listen_addrs, entry, next) {
 		TAILQ_FOREACH(ola, xconf->listen_addrs, entry)
 			if (!memcmp(&nla->sa, &ola->sa, sizeof(nla->sa)))
 				break;
@@ -423,8 +421,7 @@ merge_config(struct bgpd_config *xconf, struct bgpd_config *conf)
 			ola->reconf = RECONF_KEEP;
 	}
 	/* finally clean up the original list and remove all stale entries */
-	for (nla = TAILQ_FIRST(xconf->listen_addrs); nla != NULL; nla = next) {
-		next = TAILQ_NEXT(nla, entry);
+	TAILQ_FOREACH_SAFE(nla, xconf->listen_addrs, entry, next) {
 		if (nla->reconf == RECONF_DELETE) {
 			TAILQ_REMOVE(xconf->listen_addrs, nla, entry);
 			free(nla);
@@ -568,8 +565,7 @@ prepare_listeners(struct bgpd_config *conf)
 	int			 opt = 1;
 	int			 r = 0;
 
-	for (la = TAILQ_FIRST(conf->listen_addrs); la != NULL; la = next) {
-		next = TAILQ_NEXT(la, entry);
+	TAILQ_FOREACH_SAFE(la, conf->listen_addrs, entry, next) {
 		if (la->reconf != RECONF_REINIT)
 			continue;
 
