@@ -1,4 +1,4 @@
-/*	$OpenBSD: rde_attr.c,v 1.135 2024/09/10 09:38:45 claudio Exp $ */
+/*	$OpenBSD: rde_attr.c,v 1.137 2025/04/14 11:49:39 claudio Exp $ */
 
 /*
  * Copyright (c) 2004 Claudio Jeker <claudio@openbsd.org>
@@ -80,21 +80,20 @@ attr_optadd(struct rde_aspath *asp, uint8_t flags, uint8_t type,
 	struct attr	*a, *t;
 	void		*p;
 
-	/* known optional attributes were validated previously */
-	if ((a = attr_lookup(flags, type, data, len)) == NULL)
-		a = attr_alloc(flags, type, data, len);
-
 	/* attribute allowed only once */
 	for (l = 0; l < asp->others_len; l++) {
 		if (asp->others[l] == NULL)
 			break;
-		if (type == asp->others[l]->type) {
-			if (a->refcnt == 0)
-				attr_put(a);
+		if (type == asp->others[l]->type)
 			return (-1);
-		}
+		if (type < asp->others[l]->type)
+			break;
 	}
 
+	/* known optional attributes were validated previously */
+	if ((a = attr_lookup(flags, type, data, len)) == NULL)
+		a = attr_alloc(flags, type, data, len);
+
 	/* add attribute to the table but first bump refcnt */
 	a->refcnt++;
 	rdemem.attr_refs++;
@@ -114,7 +113,7 @@ attr_optadd(struct rde_aspath *asp, uint8_t flags, uint8_t type,
 
 	/* no empty slot found, need to realloc */
 	if (asp->others_len == UCHAR_MAX)
-		fatalx("attr_optadd: others_len overflow");
+		fatalx("attr_optadd: attribute overflow");
 
 	asp->others_len++;
 	if ((p = reallocarray(asp->others,
@@ -190,6 +189,8 @@ attr_diff(struct attr *oa, struct attr *ob)
 		return (1);
 	if (oa->len < ob->len)
 		return (-1);
+	if (oa->len == 0)
+		return (0);
 	r = memcmp(oa->data, ob->data, oa->len);
 	if (r > 0)
 		return (1);
