@@ -1,4 +1,4 @@
-/*	$OpenBSD: rde_rib.c,v 1.267 2025/01/14 12:24:23 claudio Exp $ */
+/*	$OpenBSD: rde_rib.c,v 1.272 2025/09/24 13:27:18 claudio Exp $ */
 
 /*
  * Copyright (c) 2003, 2004 Claudio Jeker <claudio@openbsd.org>
@@ -42,8 +42,8 @@ struct rib flowrib = { .id = 1, .tree = RB_INITIALIZER(&flowrib.tree) };
 struct rib_entry *rib_add(struct rib *, struct pt_entry *);
 static inline int rib_compare(const struct rib_entry *,
 			const struct rib_entry *);
-void rib_remove(struct rib_entry *);
-int rib_empty(struct rib_entry *);
+static void rib_remove(struct rib_entry *);
+static inline int rib_empty(struct rib_entry *);
 static void rib_dump_abort(uint16_t);
 
 RB_PROTOTYPE(rib_tree, rib_entry, rib_e, rib_compare);
@@ -222,7 +222,9 @@ rib_find(char *name)
 		return RIB_LOC_START;
 
 	for (id = 0; id < rib_size; id++) {
-		if (ribs[id] != NULL && !strcmp(ribs[id]->name, name))
+		/* cannot trust name to be properly terminated */
+		if (ribs[id] != NULL &&
+		    !strncmp(ribs[id]->name, name, sizeof(ribs[id]->name)))
 			return id;
 	}
 
@@ -366,13 +368,12 @@ rib_add(struct rib *rib, struct pt_entry *pte)
 		return (NULL);
 	}
 
-
 	rdemem.rib_cnt++;
 
 	return (re);
 }
 
-void
+static void
 rib_remove(struct rib_entry *re)
 {
 	if (!rib_empty(re))
@@ -391,7 +392,7 @@ rib_remove(struct rib_entry *re)
 	rdemem.rib_cnt--;
 }
 
-int
+static inline int
 rib_empty(struct rib_entry *re)
 {
 	return TAILQ_EMPTY(&re->prefix_h);
@@ -434,6 +435,8 @@ rib_dump_r(struct rib_context *ctx)
 
 	for (i = 0; re != NULL; re = next) {
 		next = RB_NEXT(rib_tree, unused, re);
+		if (rib_empty(re))
+			continue;
 		if (re->rib_id != ctx->ctx_id)
 			fatalx("%s: Unexpected RIB %u != %u.", __func__,
 			    re->rib_id, ctx->ctx_id);
@@ -492,6 +495,19 @@ rib_dump_runner(void)
 	}
 }
 
+static void
+rib_dump_free(struct rib_context *ctx)
+{
+	if (ctx->ctx_done)
+		ctx->ctx_done(ctx->ctx_arg, ctx->ctx_aid);
+	if (ctx->ctx_re && rib_empty(re_unlock(ctx->ctx_re)))
+		rib_remove(ctx->ctx_re);
+	if (ctx->ctx_p && prefix_is_dead(prefix_unlock(ctx->ctx_p)))
+		prefix_adjout_destroy(ctx->ctx_p);
+	LIST_REMOVE(ctx, entry);
+	free(ctx);
+}
+
 static void
 rib_dump_abort(uint16_t id)
 {
@@ -500,14 +516,7 @@ rib_dump_abort(uint16_t id)
 	LIST_FOREACH_SAFE(ctx, &rib_dumps, entry, next) {
 		if (id != ctx->ctx_id)
 			continue;
-		if (ctx->ctx_done)
-			ctx->ctx_done(ctx->ctx_arg, ctx->ctx_aid);
-		if (ctx->ctx_re && rib_empty(re_unlock(ctx->ctx_re)))
-			rib_remove(ctx->ctx_re);
-		if (ctx->ctx_p && prefix_is_dead(prefix_unlock(ctx->ctx_p)))
-			prefix_adjout_destroy(ctx->ctx_p);
-		LIST_REMOVE(ctx, entry);
-		free(ctx);
+		rib_dump_free(ctx);
 	}
 }
 
@@ -519,14 +528,7 @@ rib_dump_terminate(void *arg)
 	LIST_FOREACH_SAFE(ctx, &rib_dumps, entry, next) {
 		if (ctx->ctx_arg != arg)
 			continue;
-		if (ctx->ctx_done)
-			ctx->ctx_done(ctx->ctx_arg, ctx->ctx_aid);
-		if (ctx->ctx_re && rib_empty(re_unlock(ctx->ctx_re)))
-			rib_remove(ctx->ctx_re);
-		if (ctx->ctx_p && prefix_is_dead(prefix_unlock(ctx->ctx_p)))
-			prefix_adjout_destroy(ctx->ctx_p);
-		LIST_REMOVE(ctx, entry);
-		free(ctx);
+		rib_dump_free(ctx);
 	}
 }
 
@@ -1430,7 +1432,11 @@ prefix_adjout_flush_pending(struct rde_peer *peer)
 		RB_FOREACH_SAFE(p, prefix_tree, &peer->updates[aid], np) {
 			p->flags &= ~PREFIX_FLAG_UPDATE;
 			RB_REMOVE(prefix_tree, &peer->updates[aid], p);
-			peer->stats.pending_update--;
+			if (p->flags & PREFIX_FLAG_EOR) {
+				prefix_adjout_destroy(p);
+			} else {
+				peer->stats.pending_update--;
+			}
 		}
 	}
 }
