@@ -1,4 +1,4 @@
-/*	$OpenBSD: timer.c,v 1.19 2020/12/11 12:00:01 claudio Exp $ */
+/*	$OpenBSD: timer.c,v 1.20 2025/02/20 19:47:31 claudio Exp $ */
 
 /*
  * Copyright (c) 2003-2007 Henning Brauer <henning@openbsd.org>
@@ -23,19 +23,6 @@
 #include "session.h"
 #include "log.h"
 
-#define MAXIMUM(a, b)	(((a) > (b)) ? (a) : (b))
-
-time_t
-getmonotime(void)
-{
-	struct timespec	ts;
-
-	if (clock_gettime(CLOCK_MONOTONIC, &ts) != 0)
-		fatal("clock_gettime");
-
-	return (ts.tv_sec);
-}
-
 struct timer *
 timer_get(struct timer_head *th, enum Timer timer)
 {
@@ -49,34 +36,35 @@ timer_get(struct timer_head *th, enum Timer timer)
 }
 
 struct timer *
-timer_nextisdue(struct timer_head *th, time_t now)
+timer_nextisdue(struct timer_head *th, monotime_t now)
 {
 	struct timer *t;
 
 	t = TAILQ_FIRST(th);
-	if (t != NULL && t->val > 0 && t->val <= now)
+	if (t != NULL && monotime_valid(t->val) &&
+	    monotime_cmp(t->val, now) <= 0)
 		return (t);
 	return (NULL);
 }
 
-time_t
-timer_nextduein(struct timer_head *th, time_t now)
+monotime_t
+timer_nextduein(struct timer_head *th)
 {
 	struct timer *t;
 
-	if ((t = TAILQ_FIRST(th)) != NULL && t->val > 0)
-		return (MAXIMUM(t->val - now, 0));
-	return (-1);
+	if ((t = TAILQ_FIRST(th)) != NULL && monotime_valid(t->val))
+		return t->val;
+	return monotime_clear();
 }
 
 int
-timer_running(struct timer_head *th, enum Timer timer, time_t *left)
+timer_running(struct timer_head *th, enum Timer timer, monotime_t *due)
 {
 	struct timer	*t = timer_get(th, timer);
 
-	if (t != NULL && t->val > 0) {
-		if (left != NULL)
-			*left = t->val - getmonotime();
+	if (t != NULL && monotime_valid(t->val)) {
+		if (due != NULL)
+			*due = t->val;
 		return (1);
 	}
 	return (0);
@@ -87,21 +75,26 @@ timer_set(struct timer_head *th, enum Timer timer, u_int offset)
 {
 	struct timer	*t = timer_get(th, timer);
 	struct timer	*next;
+	monotime_t	 ms;
+
+	ms = monotime_from_sec(offset);
+	ms = monotime_add(ms, getmonotime());
 
 	if (t == NULL) {	/* have to create */
 		if ((t = malloc(sizeof(*t))) == NULL)
 			fatal("timer_set");
 		t->type = timer;
 	} else {
-		if (t->val == getmonotime() + (time_t)offset)
+		if (monotime_cmp(t->val, ms) == 0)
 			return;
 		TAILQ_REMOVE(th, t, entry);
 	}
 
-	t->val = getmonotime() + offset;
+	t->val = ms;
 
 	TAILQ_FOREACH(next, th, entry)
-		if (next->val == 0 || next->val > t->val)
+		if (!monotime_valid(next->val) ||
+		    monotime_cmp(next->val, t->val) > 0)
 			break;
 	if (next != NULL)
 		TAILQ_INSERT_BEFORE(next, t, entry);
@@ -115,7 +108,7 @@ timer_stop(struct timer_head *th, enum Timer timer)
 	struct timer	*t = timer_get(th, timer);
 
 	if (t != NULL) {
-		t->val = 0;
+		t->val = monotime_clear();
 		TAILQ_REMOVE(th, t, entry);
 		TAILQ_INSERT_TAIL(th, t, entry);
 	}
